#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <string.h>
#include <unistd.h>
#include "lodepng.h"
#include "font.h"

#define SCREEN_W 480
#define SCREEN_H 320
#define SCREENSZ (3 * SCREEN_W * SCREEN_H)
uint8_t       screen[SCREENSZ];

struct __attribute__( ( packed ) ) font
{
  uint16_t      height;
  char          ch_start;
  char          ch_end;
  uint16_t     *sizes;
  uint8_t      *data;
};

struct __attribute__( ( packed ) ) character
{
  uint16_t      width;
  uint16_t      offset;
};

struct font   font_8 = { 11, ' ', '~', font_8_sizes, font_8_data };
struct font   font_9 = { 13, ' ', '~', font_9_sizes, font_9_data };
struct font   font_12 = { 15, ' ', 'z', font_12_sizes, font_12_data };
struct font   font_18 = { 15, '0', 'F', font_18_sizes, font_18_data };
struct font   font_19 = { 20, '*', 'Z', font_19_sizes, font_19_data };
struct font   font_20 = { 20, ',', ':', font_20_sizes, font_20_data };
struct font   font_28 = { 28, ',', '<', font_28_sizes, font_28_data };

void
put_char( int x_coord, int y_coord, struct font *font,
    uint8_t c_char, uint16_t fg_color, uint16_t bg_color )
{
  struct character *ch;
  uint8_t      *f_data;
  int           char_off;
  uint16_t      width;
  uint8_t      *ch_data;
  uint8_t      *ch_end;
  uint8_t      *ch_ptr;
  uint8_t      *addr_char;
  uint8_t      *addr_limit;
  uint8_t      *addr_pix;
  uint8_t       fg_r, fg_g, fg_b, bg_r, bg_g, bg_b;
  unsigned char mask, pixels;
  int           height;

  bg_r = ( ( bg_color >> 11 ) & 0x1F ) << 3;
  bg_g = ( ( bg_color >> 5 ) & 0x3F ) << 2;
  bg_b = ( ( bg_color >> 0 ) & 0x1F ) << 3;
  fg_r = ( ( fg_color >> 11 ) & 0x1F ) << 3;
  fg_g = ( ( fg_color >> 5 ) & 0x3F ) << 2;
  fg_b = ( ( fg_color >> 0 ) & 0x1F ) << 3;
  height = font->height >> 3;
  if( font->height & 7 )
    height = ( font->height >> 3 ) + 1;
  ch = ( struct character * ) font->sizes;
  char_off = c_char - font->ch_start;
  ch = &ch[char_off];
  width = ch->width;
  f_data = font->data;
  if( !width || !height )
    return;
  addr_char = screen + 3 * ( SCREEN_W * y_coord + x_coord);
  ch_data = f_data + ch->offset;
  ch_end = ch_data + width;
  for( ; ch_data < ch_end; ch_data++, addr_char += 3 )
  {
    ch_ptr = ch_data;
    addr_limit = addr_char + 8 * 3 * SCREEN_W * ( height - 1 ) +
	8 * 3 * SCREEN_W;
    for( addr_pix = addr_char; addr_pix < addr_limit; )
    {
      pixels = *ch_ptr;
      ch_ptr += width;
      for( mask = 1; mask != 0; mask <<= 1 )
      {
	if( pixels & mask )
	{
	  addr_pix[0] = fg_r;
	  addr_pix[1] = fg_g;
	  addr_pix[2] = fg_b;
	}
	else
	{
	  addr_pix[0] = bg_r;
	  addr_pix[1] = bg_g;
	  addr_pix[2] = bg_b;
	}
	addr_pix += 3 * SCREEN_W;
      }
    }
  }
}

struct font  *
get_font_base( int f_size )
{
  switch ( f_size )
  {
    case 8:
      return &font_8;
    case 9:
      return &font_9;
    case 12:
      return &font_12;
    case 18:
      return &font_18;
    case 19:
      return &font_19;
    case 20:
      return &font_20;
    case 28:
      return &font_28;
    default:
      return NULL;
  }
}

int
string_width( char *str, int f_size )
{
  struct font  *font;
  struct character *ch;
  uint8_t       ch_start, ch_end;
  int           char_off, width;

  font = get_font_base( f_size );
  ch_start = font->ch_start;
  ch_end = font->ch_end;
  ch = ( struct character * ) font->sizes;
  width = 0;
  while( *str )
  {
    if( *str >= ch_start && *str <= ch_end )
    {
      char_off = *str - ch_start;
      width += ( ch + char_off )->width + 1;
    }
    str++;
  }
  return width;
}

void
put_string( int x_coord, int y_coord, int f_size, char *str,
    uint16_t fg_color, uint16_t bg_color )
{
  struct font  *font;
  struct character *ch;
  uint8_t       ch_start, ch_end;
  int           char_off;

  font = get_font_base( f_size );
  ch_start = font->ch_start;
  ch_end = font->ch_end;
  ch = ( struct character * ) font->sizes;
  while( *str )
  {
    if( *str >= ch_start && *str <= ch_end )
    {
      char_off = *str - ch_start;
      put_char( x_coord, y_coord, font, *str, fg_color, bg_color );
      x_coord += ( ch + char_off )->width + 1;
    }
    str++;
  }
}

static void
center_string( int y_coord, int f_size, char *str, uint16_t fg_color )
{
  int           x_coord = ( SCREEN_W - string_width( str, f_size ) ) / 2;

  put_string( x_coord, y_coord, f_size, str, fg_color, 0 );
}

static void
left_string( int x_coord, int y_coord, int f_size, char *str,
    uint16_t fg_color )
{
  put_string( x_coord, y_coord, f_size, str, fg_color, 0 );
}

uint8_t       device_sig[12];

void
build_screen( void )
{
  char          sprintf_buf[50];

  center_string( 30, 19, "MALAHIT-DSP", 0x7E0 );
  center_string( 62, 12, "all mode [ 50KHz - 2GHz ] radio", 0x7E0 );
  center_string( 90, 12, "version: 1.10c rev2 - 5 August 2021", 0xFFE0 );
  center_string( 114, 12, "Firmware for all genuine / clone boards", 0xFBF0 );
#define PATCH_Y    148
#define PATCH_STEP 18
  left_string( 170, PATCH_Y + 0 * PATCH_STEP, 12, "- code / screen / IDs checksums",
      0xb841 );
  left_string( 170, PATCH_Y + 1 * PATCH_STEP, 12, "- no graphic spash screen",
      0xb841 );
  left_string( 170, PATCH_Y + 2 * PATCH_STEP, 12, "- encoders interrupt code",
      0xb841 );
  left_string( 90, PATCH_Y + 2 * PATCH_STEP / 2, 12, "Patches:", 0xb841 );
  sprintf( sprintf_buf, "%04X-%04X-%04X-%04X-%04X-%04X",
      ( device_sig[0] << 8 ) | device_sig[1],
      ( device_sig[2] << 8 ) | device_sig[3],
      ( device_sig[4] << 8 ) | device_sig[5],
      ( device_sig[6] << 8 ) | device_sig[7],
      ( device_sig[8] << 8 ) | device_sig[9],
      ( device_sig[10] << 8 ) | device_sig[11] );
  center_string( 244, 12, "Your CPU ID is:", 0x7FF );
  center_string( 270, 19, sprintf_buf, 0x7FF );
}

int
main( int argc, char *argv[] )
{
  FILE         *fimg;

  if( ( fimg = fopen( "/dev/urandom", "rb" ) ) == NULL )
    return -1;
  fread( device_sig, 1, 12, fimg );
  fclose( fimg );
  build_screen(  );
  lodepng_encode24_file( "screen.png", screen, SCREEN_W, SCREEN_H );
  return 0;
}
