#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <string.h>
#include <unistd.h>
#include "font.h"

struct __attribute__( ( packed ) ) font
{
  uint16_t      height;
  char          ch_start;
  char          ch_end;
  uint16_t     *sizes;
  uint8_t      *data;
};

struct __attribute__( ( packed ) ) character
{
  uint16_t      width;
  uint16_t      offset;
};

struct font   font_8 = { 11, ' ', '~', font_8_sizes, font_8_data };
struct font   font_9 = { 13, ' ', '~', font_9_sizes, font_9_data };
struct font   font_12 = { 15, ' ', 'z', font_12_sizes, font_12_data };
struct font   font_18 = { 15, '0', 'F', font_18_sizes, font_18_data };
struct font   font_19 = { 20, '*', 'Z', font_19_sizes, font_19_data };
struct font   font_20 = { 20, ',', ':', font_20_sizes, font_20_data };
struct font   font_28 = { 28, ',', '<', font_28_sizes, font_28_data };

struct font  *
get_font_base( int f_size )
{
  switch ( f_size )
  {
    case 8:
      return &font_8;
    case 9:
      return &font_9;
    case 12:
      return &font_12;
    case 18:
      return &font_18;
    case 19:
      return &font_19;
    case 20:
      return &font_20;
    case 28:
      return &font_28;
    default:
      return NULL;
  }
}

int
main( int argc, char *argv[] )
{
  FILE         *fimg;
  struct character *ch;
  struct font  *f;
  uint8_t       b;
  uint16_t      sz[256];
  uint16_t      addr;
  int           i, j, k, fszn, bytes;
  int           fz[7] = { 8, 9, 12, 18, 19, 20, 28, };

  if( ( fimg = fopen( "font_new.h", "wt" ) ) == NULL )
    return -1;
  fprintf( fimg, "#include <stdint.h>\n\n" );
  for( k = 0; k < 7; k++ )
  {
    fszn = fz[k];
    f = get_font_base( fszn );
    bytes = ( f->height + 7 ) / 8;
    addr = 0;
    fprintf( fimg, "uint8_t       font_%d_data[] = { \n", fszn );
    for( i = 0; i <= f->ch_end - f->ch_start; i++ )
    {
      ch = ( struct character * ) f->sizes + i;
      sz[2 * i] = ch->width;
      sz[2 * i + 1] = addr;
      for( j = 0; j < bytes * ch->width; j++ )
      {
	b = f->data[ch->offset + j];
	fprintf( fimg, ( b > 9 ) ? "0x%X, " : "%d, ", b );
      }
      fprintf( fimg, "// '%c'\n", f->ch_start + i );
      addr += bytes * ch->width;
    }
    fprintf( fimg, "};\n" );
    fprintf( fimg, "uint16_t       font_%d_sizes[] = { \n", fszn );
    for( i = 0; i <= f->ch_end - f->ch_start; i++ )
    {
      fprintf( fimg, "%2d, 0x%X, // '%c'\n", sz[2 * i], sz[2 * i + 1],
	  f->ch_start + i );
    }
    fprintf( fimg, "};\n" );
  }
  fclose( fimg );
  return 0;
}
